@@ -71,6 +71,7 @@ Construct Tk::Widget 'LoginDialog';
 
 # package constants
 
+use constant A_BUTTONS => qw/ Cancel Exit Login /;
 use constant AS_DRIVERS => sort(DBI->available_drivers);
 use constant CHAR_MASK => '*';	# masking character
 use constant N_RETRY => 3;	# number of loops to attempt login
@@ -80,7 +81,7 @@ use constant S_WHATAMI => "Tk::DBI::LoginDialog";
 
 
 # --- package globals ---
-our $VERSION = '1.004';
+our $VERSION = '1.005';
 
 
 # --- package locals ---
@@ -96,10 +97,18 @@ sub ClassInit {
 
 sub CreateArgs {
 	my($class, $mw, $args) = @_;
-	my @buttons = qw/ Cancel Exit Login /;
 
-	$args->{-buttons} = [ @buttons ];
-	$args->{-default_button} = 'Login';
+	# allow button labels to be overridden in class initialisation
+	if (exists $args->{-buttons}) {
+
+		$args->{-default_button} = $args->{-buttons}->[-1]
+			unless (exists $args->{-default_button});
+
+	} else {
+		my @buttons = A_BUTTONS;
+		$args->{-buttons} = [ @buttons ];
+		$args->{-default_button} = $buttons[-1];
+	}
 
 #	printf "DEBUG args [%s]\n", Dumper($args);
 
@@ -120,6 +129,7 @@ sub Populate {
 
 	my $attribute = $self->privateData;
 	%$attribute = (
+	    buttons => undef,
 	    dbh => undef,
 	    driver => S_NULL,
 	    drivers => [ AS_DRIVERS ],
@@ -136,7 +146,9 @@ sub Populate {
 
 	$self->Advertise('LoginDialog' => $self);
 
+	$specs{-connect} = [ qw/ METHOD connect Connect /, undef ];
 	$specs{-dbh} = [ qw/ METHOD dbh Dbh /, undef ];
+	$specs{-disconnect} = [ qw/ METHOD disconnect Disconnect /, undef ];
 	$specs{-driver} = [ qw/ METHOD driver Driver /, undef ];
 	$specs{-drivers} = [ qw/ METHOD drivers Drivers /, undef ];
 	$specs{-dsn} = [ qw/ METHOD dsn dsn /, undef ];
@@ -158,6 +170,14 @@ The character or string used to hide (mask) the password.
 
 	$specs{-mask} = [ qw/ PASSIVE mask Mask /, CHAR_MASK ];
 
+=item B<-pressed>
+
+The name of the button pressed during a login sequence.
+
+=cut
+
+	$specs{-pressed} = [ qw/ PASSIVE pressed Pressed /, S_NULL ];
+
 =item B<-retry>
 
 The number of times that attempts will be made to login to the database
@@ -213,6 +233,43 @@ This callback refreshes items in the dialog as part of the B<Show> method.
 
 
 # --- private methods ---
+sub _button {
+	my $self = shift;
+	my $pressed = shift;
+	my $data = $self->privateData;
+	my %button;
+
+	if (defined $data->{'buttons'}) {
+
+		%button = %{ $data->{'buttons'} };
+	} else {
+
+		# populate button hash with button text, e.g.
+		#	{ button -> Cancel, button1 -> Exit, button2 -> Login }
+
+		for ($self->Subwidget) {
+			if ($_->class eq 'Button') {
+
+				$button{$_->name} = $_->cget('-text');
+
+       	         		$self->_log->trace(sprintf "subwidget name [%s]", $_->name);
+			}
+		}
+
+		$data->{'buttons'} = { %button };
+	}
+
+	$self->_log->trace(sprintf "button [%s]", Dumper(\%button));
+
+	for (keys %button) {
+		return $_
+			if ($button{$_} eq $pressed);
+	}
+
+	return undef;
+}
+
+
 sub _default_value {
 	my $self = shift;
 	my $attribute = shift;
@@ -248,6 +305,31 @@ sub _log {
 }
 
 
+sub _message {
+#	echo either the message passed or the DBI error string
+#	return the string.
+#
+	my $self = shift;
+	my $msg = shift;
+
+	if (defined $msg) {
+
+		$self->_log->info($msg);
+
+	} elsif (defined $DBI::errstr) {
+
+		$msg = $DBI::errstr;
+		$self->_log->logwarn($msg);
+
+	} else {
+		$msg = "WARNING unspecified DBI connect error";
+		$self->_log->logwarn($msg);
+	}
+
+	return $msg;
+}
+
+
 sub _paint {
 	my $self = shift;
 	my $name;
@@ -277,15 +359,17 @@ sub _paint {
 
 	# create all widgets first then handle geometry later
 
-#	+-------------------------+
-#	| label | BrowseEntry     |
-#	+-------------------------+
-#	| label | Entry x 3       |
-#	+-------------------------+
-#	| ROText (error)          |
-#	+-------------------------+
-#	| label (hidden, version) |
-#	+-------------------------+
+#	+---------------------+
+#	| label | BrowseEntry | driver
+#	+---------------------+
+#	| label | Entry       | dsn
+#	| label | Entry       | username
+#	| label | Entry       | password
+#	+---------------------+
+#	| ROText              | error
+#	+---------------------+
+#	| label (hidden)      | version
+#	+---------------------+
 
 	my $ld = $f->Label(-text => 'Driver', );
 	my $li = $f->Label(-textvariable => \$data->{'dsn_label'},);
@@ -336,6 +420,23 @@ Widget references for the basic credential entry widgets.
 		$e{$short} = $e;
 	}
 
+=item Name:  L_driver, Class: Label
+
+=item Name:  L_dsn, Class: Label
+
+=item Name:  L_username, Class: Label
+
+=item Name:  L_password, Class: Label
+
+Widget references of the left-most label widgets.
+
+=cut
+
+	$self->Advertise('L_driver', $ld);
+	$self->Advertise('L_dsn', $li);
+	$self->Advertise('L_username', $lu);
+	$self->Advertise('L_password', $lp);
+
 =item Name:  error, Class: ROText
 
 Widget reference of the status/error message widget.
@@ -383,36 +484,34 @@ sub cb_login {
 	my $button = shift;
 	my $data = $self->privateData;
 
-#	$self->_log->debug("DEBUG button [$button]");
+	unless (defined $button) { # Bug #108406 fix for WM event, e.g. close
 
-	if ($button eq 'Exit') {
+		$self->_log->logwarn("WARNING no action detected");
+		$self->configure('-pressed' => S_NULL);
+
+		return;
+	}
+	$self->configure('-pressed' => $button);
+
+
+	my $button_id = $self->_button($button);
+	$self->_log->trace("button_id [$button_id] button [$button]");
+
+
+	if ($button_id eq "button") {		# default 'Cancel' button
+
+	} elsif ($button_id eq "button1") {	# default 'Exit' button
 
 		$self->Callback('-exit');
 
-	} elsif ($button eq 'Cancel') {
+	} elsif ($button_id eq "button2") {	# 'default Login' button
 
-	} elsif ($button eq 'Login') {
-		$self->_log->debug("attempting to login to database");
+		my ($dbh,$msg) = $self->connect($data->{'driver'}, $data->{'dsn'}, $data->{'username'}, $data->{'password'});
 
-		my $data_source = join(':', "DBI", $data->{'driver'}, 
-			defined($data->{'dsn'}) ? $data->{'dsn'} : S_NULL
-			);
+		$data->{'dbh'} = $dbh
+			if (defined $dbh);
 
-		$self->_log->debug("data_source [$data_source]");
-
-		my $dbh = DBI->connect($data_source, $data->{'username'}, $data->{'password'});
-
-		if (defined $dbh) {
-			$data->{'dbh'} = $dbh;
-			$self->_error("Connected okay.");
-		} elsif (defined $DBI::errstr) {
-			$self->_log->logwarn($DBI::errstr);
-			$self->_error($DBI::errstr);
-		} else {
-			my $msg = "WARNING unspecified DBI connect error";
-			$self->_log->logwarn($msg);
-			$self->_error($msg);
-		}
+		$self->_error($msg);
 	} else {
 		$self->_log->logconfess("ERROR invalid action [$button]");
 	}
@@ -450,6 +549,36 @@ sub cb_populate {
 
 =over 4
 
+=item B<connect>(Driver, DSN, Username, Password)
+
+The DBI connection routine.  This does not interact with any Tk widgets
+so can be called natively, if required.  This routine is also called when
+the B<Login> button is pressed.  Returns a database handle and a string
+message indicating status of connection attempt.
+
+=cut
+
+sub connect {
+	my $self = shift;
+	$self->_log->logconfess("SYNTAX: connect(Driver, DSN, Username, Password)") unless (@_ == 4);
+	my $driver = shift; 
+	my $dsn = shift;
+	my $username = shift;
+	my $password = shift;
+
+	$self->_log->debug("attempting to login to database");
+
+	my $source = join(':', "DBI", $driver, defined($dsn) ? $dsn : S_NULL);
+
+	$self->_log->debug("source [$source]");
+
+	my $dbh = DBI->connect($source, $username, $password);
+
+	my $msg = (defined $dbh) ? $self->_message("Connected okay.") : $self->_message;
+
+	return ($dbh, $msg);
+}
+
 =item B<dbh>
 
 Returns the database handle associated with the current object.
@@ -461,6 +590,35 @@ sub dbh {
 }
 
 
+=item B<disconnect>
+
+Will call the DBI disconnection routine, using the stored handle.
+This routine is called when the B<Login> button is pressed. 
+
+=cut
+
+sub disconnect {
+	my $self = shift;
+	$self->_log->logconfess("SYNTAX: disconnect") unless (@_ == 0);
+
+	my $dbh = $self->dbh;
+	my $msg;
+
+	if (defined $dbh) {
+
+		$self->_log->debug("attempting to disconnect from database");
+
+		$dbh->disconnect;
+
+		$msg = $self->_message("Disonnected okay.");
+
+	} else {
+		$msg = $self->_message("no database connection exists");
+	}
+
+	return $msg;
+}
+
 =item B<driver> [EXPR]
 
 Set or return the B<driver> property.  For specific drivers, the label
@@ -623,7 +781,8 @@ sub login {
 
 		my $button = $self->Show;
 
-		last if (defined $self->dbh || $button =~ "Cancel");
+		last if (defined $self->dbh);
+		last if (defined $button && $button eq "Cancel");
 	} 
 
 	return $self->dbh;
@@ -696,7 +855,7 @@ __END__
 
 =head1 VERSION
 
-Build V1.004
+Build V1.005
 
 =head1 AUTHOR
 
